@@ -1,11 +1,11 @@
 /*
  * X Font Server for *.ttf Files
  *
- * $Id: xfstt.cc,v 1.1.1.1 2002/11/14 12:08:08 guillem Exp $
+ * $Id: xfstt.cc,v 1.13 2003/07/29 04:04:45 guillem Exp $
  *
  * Copyright (C) 1997-1999 Herbert Duerr
  * portions are (C) 1999 Stephen Carpenter and others
- * portions are (C) 2002 Guillem Jover
+ * portions are (C) 2002-2003 Guillem Jover
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -30,15 +30,9 @@
  */
 #define UNSTRAPLIMIT	10500U
 
-// Change these if you don't like being FHS complient
-#define TTFONTDIR	"/usr/share/fonts/truetype"
-#define TTCACHEDIR	"/var/cache/xfstt"
-
 #define TTINFO_LEAF	"ttinfo.dir"
 #define TTNAME_LEAF	"ttname.dir"
 
-#define PIDFILE		"/var/run/xfstt.pid"	// be a good little daemon
-
 #define MAXOPENFONTS	256
 #define MAXREPLYSIZE	(1 << 22)
 #define MAXFONTBUFSIZE	(1 << 24)
@@ -96,8 +90,8 @@ U16 maxLastChar = 255;
 
 static unsigned infoSize, nameSize, aliasSize;
 static char *infoBase, *nameBase, *aliasBase;
-char *fontdir = TTFONTDIR;
-char *cachedir = TTCACHEDIR;
+char *fontdir = FONTDIR;
+char *cachedir = CACHEDIR;
 char *pidfilename = PIDFILE;
 
 int defaultres = 0;
@@ -117,8 +111,8 @@ Encoding *encodings[MAXENC];
 static void
 usage(int verbose)
 {
-	printf(_("Xfstt %s, X font server for TT fonts\n"), VERSION);
-	printf(_("Usage: xfstt [[--gslist]--sync][--port portno][--unstrap]"
+	printf(_("xfstt %s, X font server for truetype fonts\n"), VERSION);
+	printf(_("Usage: xfstt [--gslist][--sync][--port portno][--unstrap]"
 		"[--user username]\n"
 		"\t\t[--dir ttfdir][--cache ttfcachedir][--pidfile pidfile]\n"
 		"\t\t[--encoding list_of_encodings][--res resolution]\n"
@@ -127,17 +121,17 @@ usage(int verbose)
 	if (!verbose)
 		return;
 
-	printf(_("\t--sync     put ttf-fonts in \"%s\" to database\n"),
-	       fontdir);
+	printf(_("\t--sync     sync database with font directory content\n"));
 	printf(_("\t--gslist   print ghostscript style ttf fontlist\n"));
 	printf(_("\t--port     change port number (default %i)\n"),
 	       default_port);
 	printf(_("\t--notcp    don't open TCP socket, use unix domain only\n"));
-	printf(_("\t--dir      change font directory (default %s)\n"),
-	       TTFONTDIR);
-	printf(_("\t--cache    change font cache directory (default %s)\n"),
-	       TTCACHEDIR);
-	printf(_("\t--pidfile  change pid file location (default %s)\n"), PIDFILE);
+	printf(_("\t--dir      change font directory\n"));
+	printf(_("\t           (default %s)\n"), FONTDIR);
+	printf(_("\t--cache    change font cache directory\n"));
+	printf(_("\t           (default %s)\n"), CACHEDIR);
+	printf(_("\t--pidfile  change pid file location\n"));
+	printf(_("\t           (default %s)\n"), PIDFILE);
 	printf(_("\t--res      force default resolution to this value\n"));
 	printf(_("\t--encoding change encoding (default %s)\n"),
 	       encodings[0]->strName);
@@ -245,9 +239,14 @@ ttSyncDir(FILE *infoFile, FILE *nameFile, char *ttdir, int gslist)
 static char *
 cachefile(const char *leafname)
 {
+	if (chdir(cachedir)) {
+		fprintf(stderr, _("xfstt: \"%s\" does not exist!\n"), cachedir);
+		return 0;
+	}
+
 	long len = strlen(cachedir) + strlen(leafname) + 2;
 	long maxlen = pathconf(cachedir, _PC_PATH_MAX);
-	
+
 	if (maxlen != -1 && len > maxlen) {
 		fputs(_("xfstt: Cache directory name is too long\n"), stderr);
 		return 0;
@@ -263,7 +262,7 @@ static int
 ttSyncAll(int gslist = 0)
 {
 	if (!gslist)
-		dprintf0("TT synching\n");
+		debug("TrueType synching\n");
 
 	if (chdir(fontdir)) {
 		fprintf(stderr, _("xfstt: \"%s\" does not exist!\n"), fontdir);
@@ -279,7 +278,7 @@ ttSyncAll(int gslist = 0)
 		delete ttinfofilename;
 		return -1;
 	}
-	
+
 	FILE *infoFile = fopen(ttinfofilename, "wb");
 	FILE *nameFile = fopen(ttnamefilename, "wb");
 	delete ttinfofilename;
@@ -386,7 +385,7 @@ listXLFDFonts(char *pattern0, int index, char *buf)
 	} else
 		mapIndex = 0;
 
-	dprintf1("match\t\"%s\"\n", buf0 + 1);
+	debug("match\t\"%s\"\n", buf0 + 1);
 
 	*buf0 = buf - buf0;
 	return *buf0 + 1;
@@ -440,7 +439,7 @@ listTTFNFonts(char *pattern, int index, char *buf)
 	strncpy(buf, (char *)&fn, sizeof(fn));
 	strncpy(buf + sizeof(fn), fontName, ttfn->nameLen);
 	buf[fn.nameLen] = 0;
-	dprintf1("ListFont \"%s\"\n", buf);
+	debug("ListFont \"%s\"\n", buf);
 
 	return fn.nameLen + 1;
 }
@@ -453,7 +452,7 @@ findFont(Font fid, int sd, int seqno)
 		if (fid == xfs->fid)
 			return xfs;
 
-	dprintf1("fid = %ld not found!\n", fid);
+	debug("fid = %ld not found!\n", fid);
 
 	if (sd) {
 		fsError reply;
@@ -471,15 +470,15 @@ static XFSFont *
 openFont(TTFont *ttFont, FontParams *fp, Rasterizer *raster,
 	 int fid, Encoding *encoding)
 {
-	dprintf3("point %d, pixel %d, res %d\n",
-		fp->point[0], fp->pixel[0], fp->resolution[0]);
+	debug("point %d, pixel %d, res %d\n",
+	      fp->point[0], fp->pixel[0], fp->resolution[0]);
 
 	if (!ttFont || ttFont->badFont())
 		return 0;
 
 	XFSFont *xfs = findFont(0, 0, 0);
 	if (!xfs) {
-		dprintf0("Too many open fonts!\n");
+		debug("Too many open fonts!\n");
 		delete ttFont;
 		return 0;
 	}
@@ -525,8 +524,8 @@ openFont(TTFont *ttFont, FontParams *fp, Rasterizer *raster,
 		fp->point[3] = (fp->pixel[3] * 72 + 36) / fp->resolution[1];
 	}
 
-	dprintf3("point %d, pixel %d, res %d\n", fp->point[0], fp->pixel[0],
-		 fp->resolution[0]);
+	debug("point %d, pixel %d, res %d\n", fp->point[0], fp->pixel[0],
+	      fp->resolution[0]);
 
 	// init rasterizer
 	raster->useTTFont(ttFont, fp->flags);
@@ -565,8 +564,8 @@ openTTFN(Rasterizer *raster, char *ttfnName, FontParams *fp, int fid)
 		return 0;
 
 	// parse attributes
-	dprintf3("point %d, pixel %d, res %d\n",
-		fp->point[0], fp->pixel[0], fp->resolution[0]);
+	debug("point %d, pixel %d, res %d\n",
+	      fp->point[0], fp->pixel[0], fp->resolution[0]);
 
 	int m_index = 0, p_index = 0, r_index = 0;
 	for (ttfnName += 2;;) {
@@ -690,9 +689,9 @@ openXLFD(Rasterizer *raster, char *xlfdName, FontParams *fp, int fid)
 	if (!encoding)
 		encoding = encodings[0];
 
-	dprintf2("\nopenXLFD(\"%s\"), %s\n", xlfdName, encoding->strName);
-	dprintf3("size %d, resx %d, resy %d\n",
-		 fp->point[0], fp->resolution[0], fp->resolution[1]);
+	debug("\nopenXLFD(\"%s\"), %s\n", xlfdName, encoding->strName);
+	debug("size %d, resx %d, resy %d\n",
+	      fp->point[0], fp->resolution[0], fp->resolution[1]);
 
 	TTFNdata* ttfn = (TTFNdata *)(infoBase + sizeof(TTFNheader));
 	// XXX: linear search should be replaced
@@ -843,7 +842,7 @@ prepare2connect(int portno)
 		s_inet.sin_family = AF_INET;
 		s_inet.sin_port = htons(portno);
 		s_inet.sin_addr.s_addr = htonl(INADDR_ANY);
-		if (bind(sd_inet,(struct sockaddr *)&s_inet, sizeof(s_inet))) {
+		if (bind(sd_inet, (struct sockaddr *)&s_inet, sizeof(s_inet))) {
 			fprintf(stderr, _("Cannot open TCPIP port %d\n"), portno);
 			fprintf(stderr, _("Better try another port!\n"));
 		}
@@ -864,7 +863,7 @@ prepare2connect(int portno)
 		sd = accept(sd_unix, (struct sockaddr *)&s_unix, &saLength);
 	else if (!noTCP && FD_ISSET(sd_inet, &sdlist))
 		sd = accept(sd_inet, (struct sockaddr *)&s_inet, &saLength);
-	dprintf2("accept(saLength = %d) = %d\n", saLength, sd);
+	debug("accept(saLength = %d) = %d\n", saLength, sd);
 
 	return sd;
 }
@@ -881,10 +880,10 @@ connecting(int sd)
 	if (i < (int)sizeof(fsConnClientPrefix))
 		return 0;
 
-	dprintf0("Connecting\n");
-	dprintf1("%s endian connection\n",
-		 (req->byteOrder == 'l') ? "little" : "big");
-	dprintf2("version %d.%d\n", req->major_version, req->minor_version);
+	debug("Connecting\n");
+	debug("%s endian connection\n",
+	      (req->byteOrder == 'l') ? "little" : "big");
+	debug("version %d.%d\n", req->major_version, req->minor_version);
 
 	if ((req->byteOrder == 'l' && (*(U32 *)req & 0xff) != 'l')
 	    || (req->byteOrder == 'B' && ((*(U32 *)req >> 24) & 0xff) != 'B')) {
@@ -936,7 +935,7 @@ fixup_bitmap(FontExtent *fe, U32 hint)
 
 	register U8 *p, *end = fe->bitmaps + fe->bmplen;
 	if ((fe->bmpFormat ^ hint) & BitmapFormatMaskByte) {
-		dprintf1("slpswap SLP=%d\n", LOGSLP);
+		debug("slpswap SLP=%d\n", LOGSLP);
 		p = fe->bitmaps;
 		switch (LOGSLP) {
 		case 3:
@@ -960,7 +959,7 @@ fixup_bitmap(FontExtent *fe, U32 hint)
 	}
 
 	if ((fe->bmpFormat ^ hint) & BitmapFormatMaskBit) {
-		dprintf0("bitswap\n");
+		debug("bitswap\n");
 		U8 map[16] = {0, 8, 4, 12, 2, 10, 6, 14,
 			      1, 9, 5, 13, 3, 11, 7, 15};
 		for (p = fe->bitmaps; p < end; ++p)
@@ -968,7 +967,7 @@ fixup_bitmap(FontExtent *fe, U32 hint)
 	}
 
 	if ((format != LOGSLP) && (hint & BitmapFormatByteOrderMask == 0)) {
-		dprintf2("fmtswap SLP=%d -> fmt=%d\n", LOGSLP, format);
+		debug("fmtswap SLP=%d -> fmt=%d\n", LOGSLP, format);
 		p = fe->bitmaps;
 		if (LOGSLP == 3 && format == 4) {
 			for (; p < end; p += 2)
@@ -1009,6 +1008,32 @@ fixup_bitmap(FontExtent *fe, U32 hint)
 	fe->bmpFormat = hint;
 }
 
+static int
+send_fserror(int sd, int seqno)
+{
+	fsError reply;
+
+	reply.type = FS_Error;
+	reply.request = FSBadLength;
+	reply.sequenceNumber = seqno;
+	reply.length = sizeof(reply) >> 2;
+
+	return write(sd, (void *)&reply, sizeof(reply));
+}
+
+static int
+check_length(int sd, int seqno, fsReq *req, int expected_size)
+{
+	if (req->length < (expected_size >> 2)) {
+		debug("packet size mismatch: %d received bytes, "
+		      "%d expected bytes\n", req->length << 2, expected_size);
+		send_fserror(sd, seqno);
+		return 0;
+	} else {
+		return 1;
+	}
+}
+
 static int
 working(int sd, Rasterizer *raster, char *replybuf)
 {
@@ -1032,17 +1057,12 @@ working(int sd, Rasterizer *raster, char *replybuf)
 		sync();
 #endif
 
-		int length = ((fsReq *)buf)->length << 2;
+		fsReq *fsreq = (fsReq *)buf;
+		int length = fsreq->length << 2;
 		if (length > MAXREQSIZE) {
-			dprintf1("request too big: %d bytes\n", length);
-
-			fsError reply;
-			reply.type = FS_Error;
-			reply.request = FSBadLength;
-			reply.sequenceNumber = seqno;
-			reply.length = sizeof(reply) >> 2;
-
-			write(sd, (void *)&reply, sizeof(reply));
+			debug("too much data: %d bytes (max=%d)\n",
+			      length, MAXREQSIZE);
+			send_fserror(sd, seqno);
 			break;
 		}
 
@@ -1064,13 +1084,13 @@ working(int sd, Rasterizer *raster, char *replybuf)
 		sync();
 #endif
 
-		switch (buf[0]) {	// request type
+		switch (fsreq->reqType) {
 		case FS_Noop:
-			dprintf0("FS_Noop\n");
+			debug("FS_Noop\n");
 			break;
 
 		case FS_ListExtensions:
-			dprintf0("FS_ListExtensions\n");
+			debug("FS_ListExtensions\n");
 			{
 			fsListExtensionsReply reply;
 			reply.type = FS_Reply;
@@ -1083,7 +1103,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			break;
 
 		case FS_QueryExtension:
-			dprintf0("FS_QueryExtension\n");
+			debug("FS_QueryExtension\n");
 			{
 			fsQueryExtensionReply reply;
 			reply.type = FS_Reply;
@@ -1103,7 +1123,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			break;
 
 		case FS_ListCatalogues:
-			dprintf0("FS_ListCatalogues\n");
+			debug("FS_ListCatalogues\n");
 			{
 			fsListCataloguesReply reply;
 			reply.type = FS_Reply;
@@ -1117,7 +1137,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			break;
 
 		case FS_SetCatalogues:	// resync font database
-			dprintf0("FS_SetCatalogues\n");
+			debug("FS_SetCatalogues\n");
 			{
 			closeTTFdb();
 			ttSyncAll();
@@ -1126,7 +1146,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			break;
 
 		case FS_GetCatalogues:
-			dprintf0("FS_GetCatalogues\n");
+			debug("FS_GetCatalogues\n");
 			{
 			fsGetCataloguesReply reply;
 			reply.type = FS_Reply;
@@ -1142,12 +1162,12 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			{
 			fsSetEventMaskReq *req = (fsSetEventMaskReq *)buf;
 			event_mask = req->event_mask;
-			dprintf1("FS_SetEventMask %04X\n", event_mask);
+			debug("FS_SetEventMask %04X\n", event_mask);
 			}
 			break;
 
 		case FS_GetEventMask:
-			dprintf1("FS_GetEventMask = %04X\n", event_mask);
+			debug("FS_GetEventMask = %04X\n", event_mask);
 			{
 			fsGetEventMaskReply reply;
 			reply.type = FS_Reply;
@@ -1160,7 +1180,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			break;
 
 		case FS_CreateAC:		// don't care
-			dprintf0("FS_CreateAC\n");
+			debug("FS_CreateAC\n");
 			{
 			fsCreateACReply reply;
 			reply.type = FS_Reply;
@@ -1174,7 +1194,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			break;
 
 		case FS_FreeAC:			// don't care
-			dprintf0("FS_FreeAC\n");
+			debug("FS_FreeAC\n");
 			{
 			fsGenericReply reply;
 			reply.type = FS_Reply;
@@ -1186,16 +1206,22 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			break;
 
 		case FS_SetAuthorization:	// don't care
-			dprintf0("FS_SetAuthorization\n");
+			debug("FS_SetAuthorization\n");
 			break;
 
 		case FS_SetResolution:
 			{
 			fsSetResolutionReq *req = (fsSetResolutionReq *)buf;
-			int numres = req->num_resolutions;	// XXX 1
+			int numres = req->num_resolutions;
+			int expected_size = numres * sz_fsResolution
+					    + sz_fsSetResolutionReq;
+
+			if (!check_length(sd, seqno, fsreq, expected_size))
+				break;
+
 			fsResolution *res = (fsResolution *)(req + 1);
 
-			dprintf1("FS_SetResolution * %d\n", numres);
+			debug("FS_SetResolution * %d\n", numres);
 			for (; --numres >= 0; ++res) {
 				if (!defaultres) {
 					fp0.resolution[0] = res->x_resolution;
@@ -1204,15 +1230,15 @@ working(int sd, Rasterizer *raster, char *replybuf)
 				res->point_size /= 10;
 				fp0.point[0] = fp0.point[1] = res->point_size;
 				fp0.point[2] = fp0.point[3] = 0;
-				dprintf3("xres = %d, yres = %d, size = %d\n",
-					 res->x_resolution, res->y_resolution,
-					 res->point_size / 10);
+				debug("xres = %d, yres = %d, size = %d\n",
+				      res->x_resolution, res->y_resolution,
+				      res->point_size / 10);
 			}
 			}
 			break;
 
 		case FS_GetResolution:
-			dprintf0("FS_GetResolution\n");
+			debug("FS_GetResolution\n");
 			{
 			struct {
 				fsGetResolutionReply s1;
@@ -1237,10 +1263,14 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			{
 			fsListFontsReq* req = (fsListFontsReq *)buf;
 			char *pattern = (char *)(req + 1);
+			int expected_size = sz_fsListFontsReq + req->nbytes;
+
+			if (!check_length(sd, seqno, fsreq, expected_size))
+				break;
 
 			pattern[req->nbytes] = 0;
-			dprintf2("FS_ListFonts \"%s\" * %ld\n",
-				 pattern, req->maxNames);
+			debug("FS_ListFonts \"%s\" * %ld\n",
+			      pattern, req->maxNames);
 
 			fsListFontsReply reply;
 			reply.type = FS_Reply;
@@ -1273,7 +1303,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 				buf += len;
 				++reply.nFonts;
 			}
-			dprintf1("Found %ld fonts\n", reply.nFonts);
+			debug("Found %ld fonts\n", reply.nFonts);
 			reply.length = (sizeof(reply) + (buf - replybuf)
 				       + 3) >> 2;
 
@@ -1289,7 +1319,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			 * The high cost is multiplied by the need to go through
 			 * different sizes and resolutions.
 			 */
-			dprintf0("FS_ListFontsWithXInfo\n");
+			debug("FS_ListFontsWithXInfo\n");
 			{
 // XFSTT_X_COMPLIANT
 #if 0
@@ -1308,8 +1338,8 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			error.sequenceNumber = seqno;
 			error.length = sizeof(error) >> 2;
 
-			dprintf2(" fsError size = %u bytes, %u ints\n",
-				 sizeof(error), error.length);
+			debug(" fsError size = %u bytes, %u ints\n",
+			      sizeof(error), error.length);
 			write(sd, (void *)&error, sizeof(error));
 #endif
 			}
@@ -1320,7 +1350,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			fsOpenBitmapFontReq *req = (fsOpenBitmapFontReq *)buf;
 			char *fontName = (char *)(req + 1) + 1;
 			fontName[*(U8 *)(req + 1)] = 0;
-			dprintf1("FS_OpenBitmapFont \"%s\"", fontName);
+			debug("FS_OpenBitmapFont \"%s\"", fontName);
 
 			raster->format = (req->format_hint >> 8) & 3;
 			if (req->format_hint & 0x0c)
@@ -1338,7 +1368,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 				reply.cachable = fsTrue;
 
 				write(sd, (void *)&reply, sizeof(reply));
-				dprintf0(" opened\n");
+				debug(" opened\n");
 			} else {
 				fsError reply;
 				reply.type = FS_Error;
@@ -1347,17 +1377,17 @@ working(int sd, Rasterizer *raster, char *replybuf)
 				reply.length = sizeof(reply) >> 2;
 
 				write(sd, (void *)&reply, sizeof(reply));
-				dprintf0(" not found\n");
+				debug(" not found\n");
 			}
-			dprintf3("fhint = %04lX, fmask = %04lX, fid = %ld\n",
-				 req->format_hint, req->format_mask, req->fid);
+			debug("fhint = %04lX, fmask = %04lX, fid = %ld\n",
+			      req->format_hint, req->format_mask, req->fid);
 			}
 			break;
 
 		case FS_QueryXInfo:
 			{
 			fsQueryXInfoReq *req = (fsQueryXInfoReq *)buf;
-			dprintf1("FS_QueryXInfo fid = %ld\n", req->id);
+			debug("FS_QueryXInfo fid = %ld\n", req->id);
 
 			struct {
 				fsQueryXInfoReply s1;
@@ -1389,12 +1419,12 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			reply.s1.font_hdr_char_range_max_char_low
 				= (U8)fi->lastChar;
 
-			dprintf2("minchar = 0x%02X%02X, ",
-				 reply.s1.font_hdr_char_range_min_char_high,
-				 reply.s1.font_hdr_char_range_min_char_low);
-			dprintf2("maxchar = 0x%02X%02X\n",
-				 reply.s1.font_hdr_char_range_max_char_high,
-				 reply.s1.font_hdr_char_range_max_char_low);
+			debug("minchar = 0x%02X%02X, ",
+			      reply.s1.font_hdr_char_range_min_char_high,
+			      reply.s1.font_hdr_char_range_min_char_low);
+			debug("maxchar = 0x%02X%02X\n",
+			      reply.s1.font_hdr_char_range_max_char_high,
+			      reply.s1.font_hdr_char_range_max_char_low);
 
 			reply.s1.font_header_draw_direction = LeftToRightDrawDirection;
 			// XXX: reply.s1.font_header_default_char_high = 0;
@@ -1416,12 +1446,12 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			reply.s1.font_header_font_ascent = fe->yWinAscent;
 			reply.s1.font_header_font_descent = fe->yWinDescent;
 
-			dprintf2("FM= (asc= %d, dsc= %d, ",
-				 fe->yAscentMax, fe->yDescentMax);
-			dprintf2("wasc= %d, wdsc= %d, ",
-				 fe->yWinAscent, fe->yWinDescent);
-			dprintf2("wmin= %d, wmax= %d)\n",
-				 fe->xAdvanceMin, fe->xAdvanceMax);
+			debug("FM= (asc= %d, dsc= %d, ",
+			      fe->yAscentMax, fe->yDescentMax);
+			debug("wasc= %d, wdsc= %d, ",
+			      fe->yWinAscent, fe->yWinDescent);
+			debug("wmin= %d, wmax= %d)\n",
+			      fe->xAdvanceMin, fe->xAdvanceMax);
 
 			// we need to have some property data, otherwise
 			// the X server complains
@@ -1440,23 +1470,33 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			break;
 
 		case FS_QueryXExtents8:
-			// convert to QueryXExtents16 request
-			{
-			fsQueryXExtents8Req *req = (fsQueryXExtents8Req *)buf;
-			U8 *p8 = (U8 *)(req + 1);
-			U16 *p16 = (U16 *)p8;
-			for (i = req->num_ranges; --i >= 0;)
-				p16[i] = htons(p8[i]);
-			}
-			// fall through
 		case FS_QueryXExtents16:
 			{
 			fsQueryXExtents16Req *req = (fsQueryXExtents16Req *)buf;
-			dprintf2("FS_QueryXExtents%s fid = %ld, ",
-				(buf[0] == FS_QueryXExtents16 ? "16" : "8"),
-				req->fid);
-			dprintf2("range=%d, nranges=%ld\n",
-				req->range, req->num_ranges);
+
+			debug("FS_QueryXExtents%s fid = %ld, ",
+			      (req->reqType == FS_QueryXExtents8 ? "8" : "16"),
+			      req->fid);
+			debug("range=%d, nranges=%ld\n",
+			      req->range, req->num_ranges);
+
+			int item_size = (req->reqType == FS_QueryXExtents8)
+				        ? 1 : 2;
+			int expected_size = sz_fsQueryXExtents8Req
+				            + req->num_ranges * item_size;
+
+			if (!check_length(sd, seqno, fsreq, expected_size))
+				break;
+
+			if (req->reqType == FS_QueryXExtents8) {
+				/*
+				 * Convert to QueryXExtents16 request
+				 */
+				U8 *p8 = (U8 *)(req + 1);
+				U16 *p16 = (U16 *)p8;
+				for (i = req->num_ranges; --i >= 0;)
+					p16[i] = htons(p8[i]);
+			}
 
 			XFSFont *xfs = findFont(req->fid, sd, seqno);
 			if (!xfs)
@@ -1475,7 +1515,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 				for (; nranges > 0; nranges -= 2, ptr += 2) {
 					ptr[0] = ntohs(ptr[0]);
 					ptr[1] = ntohs(ptr[1]);
-					dprintf2("rg %d..%d\n",ptr[0],ptr[1]);
+					debug("rg %d..%d\n",ptr[0],ptr[1]);
 					for (U16 j = ptr[0]; j <= ptr[1]; ++j)
 						(ext++)->left = j;
 				}
@@ -1514,11 +1554,11 @@ working(int sd, Rasterizer *raster, char *replybuf)
 				}
 
 #if DEBUG & 2
-				dprintf2("GM[%3d = %3d] = ", ch, glyphNo);
-				dprintf2("(l= %d, r= %d, ",
-					 ext->left, ext->right);
-				dprintf3("w= %d, a= %d, d= %d);\n",
-					 ext->width, ext->ascent, ext->descent);
+				debug("GM[%3d = %3d] = ", ch, glyphNo);
+				debug("(l= %d, r= %d, ",
+				      ext->left, ext->right);
+				debug("w= %d, a= %d, d= %d);\n",
+				      ext->width, ext->ascent, ext->descent);
 #endif
 			}
 			write(sd, (void *)&reply, sizeof(reply));
@@ -1527,22 +1567,32 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			break;
 
 		case FS_QueryXBitmaps8:
-			// convert to QueryXBitmaps16 request
-			{
-			fsQueryXBitmaps8Req *req = (fsQueryXBitmaps8Req *)buf;
-			U8 *p8 = (U8 *)(req + 1);
-			U16 *p16 = (U16 *)p8;
-			for (i = req->num_ranges; --i >= 0;)
-				p16[i] = ntohs(p8[i]);
-			}
-			// fall through
 		case FS_QueryXBitmaps16:
 			{
 			fsQueryXBitmaps16Req *req = (fsQueryXBitmaps16Req *)buf;
-			dprintf2("FS_QueryXBitmaps16 fid = %ld, fmt = %04lX\n",
-				 req->fid, req->format);
-			dprintf2("range=%d, nranges=%ld\n",
-				 req->range, req->num_ranges);
+
+			debug("FS_QueryXBitmaps16 fid = %ld, fmt = %04lX\n",
+			      req->fid, req->format);
+			debug("range=%d, nranges=%ld\n",
+			      req->range, req->num_ranges);
+
+			int item_size = (req->reqType == FS_QueryXExtents8)
+				        ? 1: 2;
+			int expected_size = sz_fsQueryXBitmaps8Req
+					    + req->num_ranges * item_size;
+
+			if (!check_length(sd, seqno, fsreq, expected_size))
+				break;
+
+			if (req->reqType == FS_QueryXBitmaps8) {
+				/*
+				 * Convert to QueryXBitmaps16 request
+				 */
+				U8 *p8 = (U8 *)(req + 1);
+				U16 *p16 = (U16 *)p8;
+				for (i = req->num_ranges; --i >= 0;)
+					p16[i] = ntohs(p8[i]);
+			}
 
 			XFSFont *xfs = findFont(req->fid, sd, seqno);
 			if (!xfs)
@@ -1563,7 +1613,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 				for (; nranges > 0; nranges -= 2, ptr += 2) {
 					ptr[0] = ntohs(ptr[0]);
 					ptr[1] = ntohs(ptr[1]);
-					dprintf2("rg %d..%d\n",ptr[0],ptr[1]);
+					debug("rg %d..%d\n",ptr[0],ptr[1]);
 					for (U16 j = ptr[0]; j <= ptr[1]; ++j)
 						(ofs++)->position = j;
 				}
@@ -1608,8 +1658,8 @@ working(int sd, Rasterizer *raster, char *replybuf)
 				ofs->position = ci->tmpofs;
 
 #if DEBUG & 2
-				dprintf3("OFS[%3d = %3d] = %ld\n",
-					 ch, glyphNo, ofs->position);
+				debug("OFS[%3d = %3d] = %ld\n",
+				      ch, glyphNo, ofs->position);
 #endif
 			}
 			reply.nbytes = bmp - bmp0;
@@ -1643,7 +1693,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 		case FS_CloseFont:
 			{
 			fsCloseReq *req = (fsCloseReq *)buf;
-			dprintf1("FS_CloseFont fid = %ld\n", req->id);
+			debug("FS_CloseFont fid = %ld\n", req->id);
 
 			XFSFont *xfs = findFont(req->id, sd, seqno);
 			if (xfs) {
@@ -1655,7 +1705,7 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			break;
 
 		default:
-			dprintf1("Unknown FS request 0x%02X !\n", buf[0]);
+			debug("Unknown FS request 0x%02X !\n", fsreq->reqType);
 			{
 			fsRequestError reply;
 			reply.type = FS_Error;
@@ -1663,14 +1713,14 @@ working(int sd, Rasterizer *raster, char *replybuf)
 			reply.sequenceNumber = seqno;
 			reply.length = sizeof(reply) >> 2;
 			reply.timestamp = 0;
-			reply.major_opcode = buf[0];
-			reply.minor_opcode = buf[1];
+			reply.major_opcode = fsreq->reqType;
+			reply.minor_opcode = fsreq->data;
 
 			write(sd, (void *)&reply, sizeof(reply));
 			}
 			break;
 		}
-	dprintf0("done.\n");
+	debug("done.\n");
 	}
 
 	return 0;
@@ -1712,6 +1762,8 @@ main(int argc, char **argv)
 	int inetdConnection = 0;
 	int portno = default_port;
 	int gslist = 0;
+	int sync_db = 0;
+	int daemon = 0;
 
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
@@ -1721,12 +1773,9 @@ main(int argc, char **argv)
 
 	for (int i = 1; i < argc; ++i) {
 		if (!strcmp(argv[i], "--gslist")) {
-			gslist = 1;
+			gslist = sync_db = 1;
 		} else if (!strcmp(argv[i], "--sync")) {
-			if (ttSyncAll(gslist) <= 0)
-				fputs(_("xfstt: sync failed\n"), stderr);
-			cleanupMem();
-			return 0;
+			sync_db = 1;
 		} else if (!strcmp(argv[i], "--port")) {
 			if (i <= argc)
 				portno = xatoi(argv[++i]);
@@ -1775,28 +1824,39 @@ main(int argc, char **argv)
 			printf(_("xfstt unstrapped: you must start X11 "
 				"with \"-deferglyphs 16\" option!\n"));
 		} else if (!strcmp(argv[i], "--daemon")) {
-			// Not sure if all this is needed
-			// but AFAIK its the right way to daemon-ize
-			if (fork())
-				_exit(0);
-			fclose(stdin);
-			fclose(stdout);
-			fclose(stderr);
-			setsid();
-			if (fork())
-				_exit(0);
+			daemon = 1;
 		} else {
 			usage(0);
 			return -1;
 		}
 	}
 
-	if (inetdConnection && multiConnection) {	// inetd
-		multiConnection = 0;
-		fprintf(stderr, _("--inetd and --multi option collission\n"));
-		exit(1); // we don't know what to do ...so exit.
+	if (sync_db) {
+		if (ttSyncAll(gslist) <= 0)
+			fputs(_("xfstt: sync failed\n"), stderr);
+		cleanupMem();
+		return 0;
 	}
-	
+
+	if (inetdConnection && multiConnection) {
+		multiConnection = 0;
+		fprintf(stderr,
+			_("xfstt: --inetd and --multi option collission\n"));
+		// we don't know what to do ...so exit.
+		return 1;
+	}
+
+	if (daemon) {
+		if (fork())
+			_exit(0);
+		fclose(stdin);
+		fclose(stdout);
+		fclose(stderr);
+		setsid();
+		if (fork())
+			_exit(0);
+	}
+
 	if (newuid == (uid_t)(-2)) {
 		newuid = getuid();
 		newgid = getgid();
@@ -1856,7 +1916,7 @@ main(int argc, char **argv)
 					shutdown(sd, 2);
 				close(sd);
 
-				dprintf0("xfstt: closing a connection\n");
+				debug("xfstt: closing a connection\n");
 				cleanupMem();
 
 				return 0;
@@ -1876,7 +1936,7 @@ main(int argc, char **argv)
 		rmdir(sockdir);
 	}
 
-	dprintf0("xfstt: server down\n");
+	debug("xfstt: server down\n");
 	cleanupMem();
 
 	return 0;
